#!/etc/ConsolePi/venv/bin/python

'''
ConsolePi API

Used for remote discovery.  When ConsolePis discover each other
this allows any of them to display options for all discovered
ConsolePis in consolepi-menu (you can connect to serail devices
attached to the remotes).

mdns and gdrive provide discovery/sync mechanisms, the API is used
to ensure the remote is reachable and that the data is current.
'''
import sys

import setproctitle
import uvicorn  # NoQA
from rich.traceback import install

sys.path.insert(0, '/etc/ConsolePi/src/pypkg')
from consolepi import config, log  # type: ignore # NoQA
from consolepi.consolepi import ConsolePi  # type: ignore # NoQA
from fastapi import FastAPI  # NoQA
# from pydantic import BaseModel  # NoQA
from time import time  # NoQA
from starlette.requests import Request  # NoQA

install(show_locals=True)

setproctitle.setproctitle("consolepi-api")


cpi = ConsolePi()
cpiexec = cpi.cpiexec
local = cpi.local
if config.power:
    if not cpiexec.wait_for_threads():
        OUTLETS = cpi.pwr.data if cpi.pwr.data else None
else:
    OUTLETS = None
user = local.user
last_update = int(time())
udev_last_update = int(time())


# class Adapters(BaseModel):
#     adapters: dict

#     class Config:
#         schema_extra = {
#             'example': [{
#                 "adapters": {
#                     "/dev/FT232R-dev": {
#                         "udev": {
#                             "by_path": "/dev/serial/by-path/platform-fd500000.pcie-pci-0000:01:00.0-usb-0:1.3.1.3:1.0-port0",
#                             "by_id": "/dev/serial/by-id/pci-FTDI_FT232R_USB_UART_AC012WBO-if00-port0",
#                             "root_dev": "false",
#                             "devname": "/dev/ttyUSB5",
#                             "devpath": "/devices/platform/scb/fd500000.pcie/pci0000:00/.../ttyUSB5/tty/ttyUSB5",
#                             "id_bus": "pci",
#                             "id_model": "FT232R_USB_UART",
#                             "id_model_from_database": "FT232 Serial (UART) IC",
#                             "id_model_id": "6001",
#                             "id_path": "platform-fd500000.pcie-pci-0000:01:00.0-usb-0:1.3.1.3:1.0",
#                             "id_pci_class_from_database": "Serial bus controller",
#                             "id_pci_subclass_from_database": "USB controller",
#                             "id_serial": "FTDI_FT232R_USB_UART_AB123456",
#                             "id_serial_short": "AB123456",
#                             "id_type": "generic",
#                             "id_usb_driver": "ftdi_sio",
#                             "id_usb_interfaces": ":ffffff:",
#                             "id_ifnum": "00",
#                             "id_vendor": "FTDI",
#                             "id_vendor_from_database": "Future Technology Devices International, Ltd",
#                             "id_vendor_id": "0403",
#                             "subsystem": "tty",
#                             "time_since_init": "2:09:35.455413 as of 03/29/20 01:03:48 AM CDT"
#                         },
#                         "outlets": [],
#                         "config": {
#                             "port": 7013,
#                             "baud": 9600,
#                             "dbits": 8,
#                             "parity": "n",
#                             "flow": "n",
#                             "sbits": 1,
#                             "logfile": "null",
#                             "log_ptr": 'null',
#                             "cmd": "picocom /dev/FT232R-dev --baud 9600 --flow n --databits 8 --parity n --stopbits 1",
#                             "line": "7013:telnet:0:/dev/FT232R-dev:9600 8DATABITS NONE 1STOPBIT  banner "
#                         }
#                     }
#                 }
#             }]
#         }


app = FastAPI(title='ConsolePi.API',
              docs_url='/api/docs',
              redoc_url="/api/redoc",
              openapi_url='/api/openapi/openapi.json'
              )


def log_request(request: Request, route: str):
    log.info('[NEW API RQST IN] {} Requesting -- {} -- Data via API'.format(request.client.host, route))


#  -- Haven't yet cracked the code on properly updating swagger-ui with examples and schema --
# @app.get('/api/v1.0/adapters', responses={200: {'model': Adapters}})
@app.get('/api/v1.0/adapters')
async def adapters(request: Request, refresh: bool = False):
    global last_update
    time_upd = True if int(time()) - last_update > 20 else False
    log_request(request, f'adapters Update based on Time {time_upd}, Update based on query param {refresh}')
    # if data has been refreshed in the last 20 seconds trust it is valid
    # prevents multiple simul calls to get_adapters after mdns_refresh and
    # subsequent API calls from all other ConsolePi on the network
    if refresh or int(time()) - last_update > 20:
        config.ser2net_conf = config.get_ser2net()
        local.adapters = local.build_adapter_dict(refresh=True)
        last_update = int(time())
    return {'adapters': local.adapters}


@app.get('/api/v1.0/adapters/udev/{adapter}')
async def udev(request: Request, adapter: str = None):
    log_request(request, f'fetching udev details for {adapter}')
    return {adapter: local.udev_adapters.get(f'/dev/{adapter}')}


@app.get('/api/v1.0/remotes')
def remotes(request: Request):
    log_request(request, 'remotes')
    return {'remotes': config.get_remotes_from_file()}


@app.get('/api/v1.0/interfaces')
def get_ifaces(request: Request):
    log_request(request, 'ifaces')
    local.interfaces = local.get_if_info()
    return {'interfaces': local.interfaces}

# removing due to fastapi issue #894, outlets method was not being used for anything currently so disabling for now
# @app.get('/api/v1.0/outlets')
# def get_outlets(request: Request, outlets=OUTLETS):
#     log_request(request, 'outlets')
#     # -- Collect Outlet Details remove sensitive data --
#     if outlets:
#         outlets = cpi.pwr.pwr_get_outlets()
#         cpiexec.wait_for_threads()
#         if outlets and 'defined' in outlets:
#             outlets['defined'] = {p: {k: v for k, v in outlets['defined'][p].items()
#                                   if k not in ['username', 'password']} for p in outlets['defined']}

#     return outlets


@app.get('/api/v1.0/details')
def get_details(request: Request):
    log_request(request, 'details')
    global last_update
    if int(time()) - last_update > 20:
        local.data = local.build_local_dict(refresh=True)
        last_update = int(time())
    return local.data


if __name__ == "__main__":
    uvicorn.run(app, host="0.0.0.0", port=config.api_port, log_level="info")



#!/etc/ConsolePi/venv/bin/python3

import json
import logging
import os
import requests  # NoQA
from consolepi.utils import Utils  # type: ignore # NoQA

try:
    import better_exceptions  # type: ignore
    better_exceptions.MAX_LENGTH = None
    os.environ['BETTER_EXCEPTIONS'] = '1'
except ImportError:
    pass

LOG_FILE = '/var/log/ConsolePi/consolepi.log'


# For working in vscode, vscode no longer provides full path
if os.environ.get("TERM_PROGRAM"):
    os.environ["PATH"] = f'{os.environ["PATH"]}:/usr/local/sbin:/usr/sbin'


class Response():
    def __init__(self, ok: bool, output=None, error=None, status_code=None, state=None, do_json=False, **kwargs):
        self.ok = ok
        self.text = output
        self.error = error
        self.state = state
        self.status_code = status_code
        if 'json' in kwargs:
            self.json = kwargs['json']
        else:
            self.json = None if not do_json else json.dumps(output)


class ConsolePiAction():
    def __init__(self, *args, function=None, callback=None, calling_menu=None, update_attribute=None,
                 spin=False, confirm=False, **kwargs):
        self.function = function
        self.args = args
        self.callback = callback
        self.calling_menu = calling_menu
        self.update_attribute = update_attribute
        self.spin = spin
        self.confirm = confirm
        self.kwargs = kwargs
        self.available = self.__dict__.keys()


class ConsolePiLog:
    def __init__(self, log_file, debug=False):
        self.error_msgs = []
        self.DEBUG = debug
        self.verbose = False
        self.log_file = log_file
        self._log = self.get_logger()
        self.name = self._log.name

    def get_logger(self):
        '''Return custom log object.'''
        # fmtStr = "%(asctime)s [%(module)s:%(funcName)s:%(lineno)d:%(process)d][%(levelname)s]: %(message)s"
        fmtStr = "%(asctime)s [%(process)d][%(levelname)s]: %(message)s"
        dateStr = "%m/%d/%Y %I:%M:%S %p"
        logging.basicConfig(filename=self.log_file,
                            level=logging.DEBUG if self.DEBUG else logging.INFO,
                            format=fmtStr,
                            datefmt=dateStr)
        return logging.getLogger('ConsolePi')

    def log_print(self, msgs, log=False, show=True, level='info', *args, **kwargs):
        msgs = [msgs] if not isinstance(msgs, list) else msgs
        _msgs = []
        _logged = []
        for i in msgs:
            i = i if isinstance(i, str) else str(i)
            if log and i not in _logged:
                getattr(self._log, level)(i)
                _logged.append(i)
            if '\n' in i:
                _msgs += i.replace('\t', '').replace('\r', '').split('\n')
            elif i.startswith('[') and ']' in i:
                _msgs.append(i.split(']', 1)[1].replace('\t', '').replace('\r', ''))
            else:
                _msgs.append(i.replace('\t', '').replace('\r', '').strip())

        msgs = []
        [msgs.append(i) for i in _msgs
            if i and i not in msgs and i not in self.error_msgs]

        if show:
            self.error_msgs += msgs

    def show(self, msgs, log=False, show=True, *args, **kwargs):
        self.log_print(msgs, show=show, log=log, *args, **kwargs)

    def debug(self, msgs, log=True, show=False, *args, **kwargs):
        self.log_print(msgs, log=log, show=show, level='debug', *args, **kwargs)

    # -- more verbose debugging - primarily to get json dumps
    # -- set verbose_debug: true in OVERRIDES to enable
    def debugv(self, msgs, log=True, show=False, *args, **kwargs):
        if self.DEBUG and self.verbose:
            self.log_print(msgs, log=log, show=show, level='debug', *args, **kwargs)

    def info(self, msgs, log=True, show=False, *args, **kwargs):
        self.log_print(msgs, log=log, show=show, *args, **kwargs)

    def warning(self, msgs, log=True, show=False, *args, **kwargs):
        self.log_print(msgs, log=log, show=show, level='warning', *args, **kwargs)

    def error(self, msgs, log=True, show=False, *args, **kwargs):
        self.log_print(msgs, log=log, show=show, level='error', *args, **kwargs)

    def exception(self, msgs, log=True, show=False, *args, **kwargs):
        self.log_print(msgs, log=log, show=show, level='exception', *args, **kwargs)

    def critical(self, msgs, log=True, show=False, *args, **kwargs):
        self.log_print(msgs, log=log, show=show, level='critical', *args, **kwargs)

    def fatal(self, msgs, log=True, show=False, *args, **kwargs):
        self.log_print(msgs, log=log, show=show, level='fatal', *args, **kwargs)

    def setLevel(self, level):
        getattr(self._log, 'setLevel')(level)

    def clear(self):
        self.error_msgs = []


utils = Utils()

log = ConsolePiLog(LOG_FILE)

from consolepi.config import Config  # type: ignore # NoQA

config = Config()  # NoQA
if config.debug:
    log.setLevel(logging.DEBUG)

if config.ovrd.get('verbose_debug'):
    log.verbose = True



#!/etc/ConsolePi/venv/bin/python3

import json
import threading
import time
from typing import Any, Dict, List, Tuple, Union

try:
    import RPi.GPIO as GPIO
    is_rpi = True
except (RuntimeError, ModuleNotFoundError, ImportError):
    is_rpi = False

from consolepi import log, config, requests, utils  # type: ignore
from consolepi.power import DLI  # type: ignore

TIMING = False


class ConsolePiPowerException(Exception):
    pass


class Outlets:
    def __init__(self):
        if is_rpi:
            GPIO.setmode(GPIO.BCM)
            GPIO.setwarnings(False)
        self._dli = {}

        # Some convenience Bools used by menu to determine what options to display
        self.dli_exists = True if 'dli' in config.outlet_types or config.do_dli_menu else False
        self.tasmota_exists = True if 'tasmota' in config.outlet_types else False
        self.esphome_exists = True if 'esphome' in config.outlet_types else False  # TODO Future
        self.gpio_exists = True if 'gpio' in config.outlet_types else False
        self.linked_exists = True if config.linked_exists else False
        if self.dli_exists or self.tasmota_exists or self.esphome_exists or self.gpio_exists:
            self.outlets_exists = True
        else:
            self.outlets_exists = False

        self.data: Dict[str, Any] = config.outlets

        if config.power:
            self.pwr_start_update_threads()

    def linked(self):
        pass

    def do_tasmota_cmd(self, address, command=None):
        '''
        Perform Operation on Tasmota outlet:
        params:
        address: IP or resolvable hostname
        command:
            True | 'ON': power the outlet on
            False | 'OFF': power the outlet off
            'Toggle': Toggle the outlet
            'cycle': Cycle Power on outlets that are powered On

        TODO: Right now this method does not verify if port is currently in an ON state
            before allowing 'cycle', resulting in it powering on the port consolepi-menu
            verifies status before allowing the command, but *may* be that other outlet
            are handled by this library.. check & make consistent
        TODO: remove int returns and re-factor all returns to use a return class (like requests)
        '''
        # sub to make the api call to the tasmota device
        def tasmota_req(**kwargs):
            querystring = kwargs['querystring']
            try:
                response = requests.request("GET", url, headers=headers, params=querystring, timeout=config.so_timeout)
                if response.status_code == 200:
                    if json.loads(response.text)['POWER'] == 'ON':
                        _response = True
                    elif json.loads(response.text)['POWER'] == 'OFF':
                        _response = False
                    else:
                        _response = 'invalid state returned {}'.format(response.text)
                else:
                    _response = '[{}] error returned {}'.format(response.status_code, response.text)
            except (requests.exceptions.Timeout, requests.exceptions.ConnectionError):
                _response = 'Unreachable'
            except requests.exceptions.RequestException as e:
                log.debug(f"[tasmota_req] {url.replace('http://', '').replace('https://', '').split('/')[0]} Exception: {e}")
                _response = 'Unreachable ~ hit catchall exception handler'  # determine if other exceptions types are possible
            return _response
        # -------- END SUB --------

        url = 'http://' + address + '/cm'
        headers = {
            'Cache-Control': "no-cache",
            'Connection': "keep-alive",
            'cache-control': "no-cache"
        }

        querystring = {"cmnd": "Power"}
        cycle = False
        if command is not None:
            if isinstance(command, bool):
                command = 'ON' if command else 'OFF'

            command = command.upper()
            if command in ['ON', 'OFF', 'TOGGLE']:
                querystring = {"cmnd": f"Power {command}"}
            elif command == 'CYCLE':
                cur_state = tasmota_req(querystring={"cmnd": "Power"})
                if isinstance(cur_state, bool):
                    if cur_state:
                        querystring = {"cmnd": "Power OFF"}
                        cycle = True
                    else:
                        return False  # return False for cycle if outlet was off indicating state is invalid for cycle
                else:
                    return cur_state  # an error occurred getting current state
            else:
                raise ConsolePiPowerException(f'Invalid Type {type(command)} passed to do_tasmota_cmd')

        # -- // Send Request to TASMOTA \\ --
        r = tasmota_req(querystring=querystring)
        if cycle:
            if isinstance(r, bool):
                if not r:
                    time.sleep(config.cycle_time)
                    r = tasmota_req(querystring={"cmnd": "Power ON"})
                else:
                    return 'Unexpected response, port returned on state expected off'
        return r

    def do_esphome_cmd(self, address, relay_id, command=None):
        '''Perform Operation on espHome outlets.

        Arguments:
            address {str} -- ip of fqdn of espHome outlet
            relay_id {str} -- The id of the relay/port

        Keyword Arguments:
            command {Bool|str|None} -- The command to perform on the outlet (default: {None})
                None (Default): get current state of outlet
                True | 'ON':    power the outlet on
                False | 'OFF':  power the outlet off
                'Toggle':       Toggle the outlet
                'cycle':        Cycle Power on outlets that are powered On

        Returns:
            Bool or str -- Bool indicating state of outlet after operation or str with error text
        '''
        def esphome_req(*args, command=command):
            '''sub function to perform operation on outlet
            '''
            try:
                method = "GET" if command is None else "POST"
                response = requests.request(method, url=url,
                                            headers=headers, timeout=config.so_timeout)
                if response.status_code == 200:
                    if command is None:
                        _response = response.json().get('value')
                    else:
                        # espHome returns status Code 200 with no content for /toggle
                        if command in ['toggle', 'cycle']:
                            _response = not cur_state
                        else:
                            _response = command  # success return bool command
                else:
                    _response = '[{}] error returned {}'.format(response.status_code, response.text)
            except (requests.exceptions.Timeout, requests.exceptions.ConnectionError):
                _response = 'Unreachable'
            except requests.exceptions.RequestException as e:
                log.debug(f"[esphome_req] {url.replace('http://', '').replace('https://', '').split('/')[0]} Exception: {e}")
                _response = 'Unreachable'  # So I can determine if other exceptions types are possible when unreachable
            return _response
        # -------- END SUB --------

        url = status_url = 'http://' + address + '/switch/' + str(relay_id)
        headers = {
            'Cache-Control': "no-cache",
            'Connection': "keep-alive",
            'cache-control': "no-cache"
        }
        # -- Get initial State of Outlet --
        cur_state = esphome_req(command=None)

        cycle = False
        if command is None:
            return cur_state
        elif isinstance(command, bool):
            if command:  # Turn On Outlet
                url = status_url + '/turn_on'
                if cur_state is True:
                    return cur_state
            else:  # Turn Off Outlet
                url = status_url + '/turn_off'
                if cur_state is False:
                    return cur_state
        elif command in ['toggle', 'cycle']:
            url = status_url + '/toggle'
            if command == 'cycle':
                if cur_state:
                    cycle = True
                else:
                    return False  # Cycle invalid for outlets in off state
        else:
            return f'[PWR-ESP] DEV Note: Invalid command \'{command}\' passed to func'

        # -- // Send Request to esphome \\ --
        r = esphome_req(command=command)
        if isinstance(r, bool):
            if cycle:
                if r is False:
                    cur_state = False
                    time.sleep(config.cycle_time)
                    r = esphome_req()
                else:
                    return '[PWR-ESP] Unexpected response, port returned on state expected off'
        return r

    def load_dli(self, address, username, password):
        '''
        Returns instace of DLI class

        Response: tuple
            DLI-class-object, Bool: True if class was previously instantiated ~ needs update
                                    False indicating class was just instantiated ~ data is fresh
            i.e.: (<dli object>, True)
        '''
        if not self._dli.get(address):
            try:
                self._dli[address] = DLI(address, username, password, timeout=config.dli_timeout, log=log)
            except Exception as e:
                log.warning(f"[PWR-DLI] DLI @ {address} is now unreachable {e.__class__.__name__}", show=True)
                return None, None
            # --// Return Pass or fail based on reachability \\--
            if self._dli[address].reachable:
                return self._dli[address], False
            else:
                return None, None

        # --// DLI Already Loaded \\--
        else:
            return self._dli[address], True

    def pwr_start_update_threads(self, upd_linked: bool = False, failures: Dict[str, Any] = {}, t_name: str = 'init'):
        kwargs = {'upd_linked': upd_linked, 'failures': failures}
        outlets = self.data.get('defined', {})
        if not failures:
            if 'failures' in outlets:
                failures = outlets['failures']
        if failures:  # re-attempt connection to failed power controllers on refresh
            outlets = {**outlets, **failures}
            failures = {}

        # this shouldn't happen, but prevents spawning multiple updates for same outlet
        _outlets = outlets.copy()
        if _outlets is not None:
            for k in _outlets:
                found = False
                for t in threading.enumerate():
                    if t.name == f'{t_name}_pwr_{k}':
                        found = True
                        break
                if not found:
                    threading.Thread(target=self.pwr_get_outlets, args=[{k: _outlets[k]}],
                                     kwargs=kwargs, name=t_name + '_pwr_' + k).start()

    def update_linked_devs(self, outlet: Dict[str, Any]) -> Tuple[Dict[str, Any], List[Union[str, int]]]:
        '''Update linked devs for dli outlets if they exist

        Params:
            dict -- the outlet dict to be updated

        Returns:
            tuple -- 0: dict: updated outlet dict or same dict if no linked_devs
                     1: list: list of all ports linked on this dli (used to initiate query against the dli)
        '''
        this_dli = self._dli.get(outlet['address'])
        _p = []
        if outlet.get('linked_devs'):
            for dev in outlet['linked_devs']:
                if isinstance(outlet['linked_devs'][dev], list):
                    [_p.append(int(_)) for _ in outlet['linked_devs'][dev] if int(_) not in _p]
                elif int(outlet['linked_devs'][dev]) not in _p:
                    _p.append(int(outlet['linked_devs'][dev]))

            if isinstance(_p, int):
                outlet['is_on'] = {_p: this_dli.outlets[_p]}
            else:
                outlet['is_on'] = {port: this_dli.outlets[port] for port in _p}

        return outlet, _p

    def dli_close_all(self, dlis=None):
        '''Close Connection to any connected dli Web Power Switches

        Arguments:
            dlis {dict} -- dict of dli objects
        '''
        dlis = self._dli if not dlis else dlis
        for address in dlis:
            if dlis[address].dli:
                if getattr(dlis[address], 'rest'):
                    threading.Thread(target=dlis[address].dli.close).start()
                else:
                    threading.Thread(target=dlis[address].dli.session.close).start()

    def pwr_get_outlets(self, outlet_data: Dict[str, Any] = {}, upd_linked: bool = False, failures: Dict[str, Any] = {}) -> Dict[str, Any]:
        '''Get Details for Outlets defined in ConsolePi.yaml power section

        On Menu Launch this method is called in parallel (threaded) for each outlet
        On Refresh all outlets are passed to the method

        params: - All Optional
            outlet_data:dict, The outlets that need to be updated, if not provided will get all outlets defined in ConsolePi.yaml
            upd_linked:Bool, If True will update just the linked ports, False is for dli and will update
                all ports for the dli.
            failures:dict: when refreshing outlets pass in previous failures so they can be re-tried
        '''
        # re-attempt connection to failed power controllers on refresh
        log.debug(f"[PWR VRFY (pwr_get_outlets)] Processing {', '.join(outlet_data.keys())}")
        if not failures:
            failures = outlet_data.get('failures', {}) if outlet_data.get('failures') else self.data.get('failures', {})

        outlet_data = self.data.get('defined', {}) if not outlet_data else outlet_data
        if failures:
            outlet_data = {**outlet_data, **failures}
            failures = {}

        dli_power = self.data.get('dli_power', {})
        esp_power = self.data.get('esp_power', {})

        for k in outlet_data:
            outlet = outlet_data[k]
            _start = time.perf_counter()

            # -- // GPIO \\ --
            if outlet['type'].upper() == 'GPIO':
                if not is_rpi:
                    log.warning('GPIO Outlet Defined, GPIO Only Supported on RPi - ignored', show=True)
                    continue
                noff = True if 'noff' not in outlet else outlet['noff']
                GPIO.setup(outlet['address'], GPIO.OUT)
                outlet_data[k]['is_on'] = bool(GPIO.input(outlet['address'])) if noff \
                    else not bool(GPIO.input(outlet['address']))

            # -- // tasmota \\ --
            elif outlet['type'] == 'tasmota':
                response = self.do_tasmota_cmd(outlet['address'])
                outlet['is_on'] = response
                if response not in [0, 1, True, False]:
                    failures[k] = outlet_data[k]
                    failures[k]['error'] = f'[PWR-TASMOTA] {k}:{failures[k]["address"]} {response} - Removed'
                    log.warning(failures[k]['error'], show=True)

            # -- // esphome \\ --
            elif outlet['type'] == 'esphome':
                # TODO have do_esphome accept list, slice, or str for one or multiple relays
                relays = utils.listify(outlet.get('relays', k))  # if they have not specified the relay try name of outlet
                outlet['is_on'] = {}
                esp_ok = True
                for r in relays:
                    response = self.do_esphome_cmd(outlet['address'], r)
                    outlet['is_on'][r] = {'state': response, 'name': r}
                    if response not in [True, False]:
                        failures[k] = outlet_data[k]
                        failures[k]['error'] = f'[PWR-ESP] {k}:{r} ({failures[k]["address"]}) {response} - Removed'
                        log.warning(failures[k]['error'], show=True)
                        esp_ok = False
                        break

                # add multi-port esp_outlets to dli_menu, unless all outlets are linked anyway
                # if esp is 8 ports add it to dli regardless (dli are 8 and they get that treatment)
                if esp_ok and len(relays) > 1 or (esp_ok and len(relays) == 8):
                    no_linkage_relays = [r for r in relays if f"'{r}'" not in str(outlet_data[k]["linked_devs"])]
                    if no_linkage_relays:
                        esp_power[outlet_data[k]["address"]] = outlet_data[k]["is_on"]

            # -- // dli \\ --
            elif outlet['type'].lower() == 'dli':
                if TIMING:
                    dbg_line = '------------------------ // NOW PROCESSING {} \\\\ ------------------------'.format(k)
                    print('\n{}'.format('=' * len(dbg_line)))
                    print('{}\n{}\n{}'.format(dbg_line, outlet_data[k], '-' * len(dbg_line)))
                    print('{}'.format('=' * len(dbg_line)))

                # -- // VALIDATE CONFIG FILE DATA FOR DLI \\ --
                all_good = True  # initial value
                for _ in ['address', 'username', 'password']:
                    if not outlet.get(_):
                        all_good = False
                        failures[k] = outlet_data[k]
                        failures[k]['error'] = f'[PWR-DLI {k}] {_} missing from {failures[k]["address"]} ' \
                            'configuration - skipping'
                        log.error(f'[PWR-DLI {k}] {_} missing from {failures[k]["address"]} '
                                  'configuration - skipping', show=True)
                        break
                if not all_good:
                    continue

                (this_dli, _update) = self.load_dli(outlet['address'], outlet['username'], outlet['password'])
                if this_dli is None or this_dli.dli is None:
                    failures[k] = outlet_data[k]
                    failures[k]['error'] = '[PWR-DLI {}] {} Unreachable - Removed'.format(k, failures[k]['address'])
                    log.warning(f"[PWR-DLI {k}] {failures[k]['address']} Unreachable - Removed", show=True)
                else:
                    if TIMING:
                        xstart = time.perf_counter()
                        print('this_dli.outlets: {} {}'.format(this_dli.outlets, 'update' if _update else 'init'))
                        print(json.dumps(dli_power, indent=4, sort_keys=True))

                    # upd_linked is for faster update in power menu only refreshes data for linked ports vs entire dli
                    if upd_linked and self.data['dli_power'].get(outlet['address']):
                        if outlet.get('linked_devs'):
                            (outlet, _p) = self.update_linked_devs(outlet)
                            if k in outlet_data:
                                outlet_data[k]['is_on'] = this_dli[_p]
                            else:
                                log.error(f'[PWR GET_OUTLETS] {k} appears to be unreachable')

                            # TODO not actually using the error returned this turned into a hot mess
                            if isinstance(outlet['is_on'], dict) and not outlet['is_on']:
                                all_good = False
                            # update dli_power for the refreshed / linked ports
                            else:
                                for _ in outlet['is_on']:
                                    dli_power[outlet['address']][_] = outlet['is_on'][_]
                    else:
                        if _update:
                            dli_power[outlet['address']] = this_dli.get_dli_outlets()  # data may not be fresh trigger dli update

                            # handle error connecting to dli during refresh - when connect worked on menu launch
                            if not dli_power[outlet['address']]:
                                failures[k] = outlet_data[k]
                                failures[k]['error'] = f"[PWR-DLI] {k} {failures[k]['address']} Unreachable - Removed"
                                log.warning(f'[PWR-DLI {k}] {failures[k]["address"]} Unreachable - Removed',
                                            show=True)
                                continue
                        else:  # dli was just instantiated data is fresh no need to update
                            dli_power[outlet['address']] = this_dli.outlets

                        if outlet.get('linked_devs'):
                            (outlet, _p) = self.update_linked_devs(outlet)

                if TIMING:
                    print('[TIMING] this_dli.outlets: {}'.format(time.perf_counter() - xstart))  # type: ignore

            log.debug(f"{outlet['type'].lower()} {k} Updated. Elapsed Time(secs): {time.perf_counter() - _start}")
            # -- END for LOOP for k in outlet_data --

        # Move failed outlets from the keys that populate the menu to the 'failures' key
        # failures are displayed in the footer section of the menu, then re-tried on refresh
        for _dev in failures.copy():
            if outlet_data.get(_dev):
                del outlet_data[_dev]
            if self.data['defined'].get(_dev):
                del self.data['defined'][_dev]
            if failures[_dev]['address'] in dli_power:
                del dli_power[failures[_dev]['address']]
            self.data['failures'][_dev] = failures[_dev]

        # restore outlets that failed on menu launch but found reachable during refresh
        for _dev in outlet_data.copy():
            if _dev not in self.data['defined']:
                self.data['defined'][_dev] = outlet_data[_dev]
            if _dev in self.data['failures']:
                del self.data['failures'][_dev]

        self.data['dli_power'] = dli_power
        self.data['esp_power'] = esp_power

        log.debug(f"[PWR VRFY (pwr_get_outlets)] Done Processing {', '.join(outlet_data.keys())}")
        return self.data

    def pwr_toggle(self, pwr_type, address, desired_state=None, port=None, noff=True, noconfirm=False):
        '''Toggle Power On the specified port

        args:
            pwr_type: valid types = 'dli', 'tasmota', 'GPIO' (not case sensitive)
            address: for dli and tasmota: str - ip or fqdn
        kwargs:
            desired_state: bool The State True|False (True = ON) you want the outlet to be in
                if not provided the method will query the current state of the port and set desired_state to the inverse
            port: Only required for dli: can be type str | int | list.
                valid:
                    int: representing the dli outlet #
                    list: list of outlets(int) to perform operation on
                    str: 'all' ~ to perform operation on all outlets
            noff: Bool, default: True.  = normally off, only applies to GPIO based outlets.
                If an outlet is normally off (True) = the relay/outlet is off if no power is applied via GPIO
                Setting noff to False flips the ON/OFF evaluation so the menu will show the port is ON when no power is applied.

        returns:
            Bool representing resulting port state (True = ON)
        '''
        # --// REMOVE ONCE VERIFIED \\--
        # if isinstance(desired_state, str):  # menu should be passing in True/False no on off now. can remove once that's verified
        #     desired_state = False if desired_state.lower() == 'off' else True
        #     print('\ndev_note: pwr_toggle passed str not bool for desired_state check calling function {}'.format(desired_state))
        #     time.sleep(5)

        # -- // Toggle dli web power switch port \\ --
        if pwr_type.lower() == 'dli':
            if port is not None:
                return self._dli[address].toggle(port, toState=desired_state)

        # -- // Toggle GPIO port \\ --
        elif pwr_type.upper() == 'GPIO':
            gpio = address
            # get current state and determine inverse if toggle called with no desired_state specified
            if desired_state is None:
                cur_state = bool(GPIO.input(gpio)) if noff else not bool(GPIO.input(gpio))
                desired_state = not cur_state
            if desired_state:
                GPIO.output(gpio, int(noff))
            else:
                GPIO.output(gpio, int(not noff))
            return bool(GPIO.input(gpio)) if noff else not bool(GPIO.input(gpio))

        # -- // Toggle TASMOTA port \\ --
        elif pwr_type.lower() == 'tasmota':
            if desired_state is None:
                desired_state = not self.do_tasmota_cmd(address)
            return self.do_tasmota_cmd(address, desired_state)

        # -- // Toggle espHome port \\ --
        elif pwr_type.lower() == 'esphome':
            if desired_state is None:
                desired_state = not self.do_esphome_cmd(address, port)
            return self.do_esphome_cmd(address, port, desired_state)

        else:
            raise Exception('pwr_toggle: Invalid type ({}) or no name provided'.format(pwr_type))

    def pwr_cycle(self, pwr_type, address, port=None, noff=True):
        '''returns Bool True = Power Cycle success, False Not performed Outlet OFF

        TODO Check error handling if unreachable
        '''
        pwr_type = pwr_type.lower()
        # --// CYCLE DLI PORT \\--
        if pwr_type == 'dli':
            if port is not None:
                return self._dli[address].cycle(port)
            else:
                raise ConsolePiPowerException('pwr_cycle: port must be provided for outlet type dli')

        # --// CYCLE GPIO PORT \\--
        elif pwr_type == 'gpio':
            gpio = address
            # normally off states are normal 0:off, 1:on - if not normally off it's reversed 0:on 1:off
            cur_state = GPIO.input(gpio) if noff else not GPIO.input(gpio)
            if cur_state:
                GPIO.output(gpio, int(not noff))
                time.sleep(config.cycle_time)
                GPIO.output(gpio, int(noff))
                response = bool(GPIO.input(gpio))
                return response if noff else not response
            else:
                return False  # Cycle is not valid on ports that are already off

        # --// CYCLE TASMOTA PORT \\--
        elif pwr_type == 'tasmota':
            # response = self.do_tasmota_cmd(address)  # this logic is handled in the function now
            # if response:  # Only Cycle if outlet is currently ON
            return self.do_tasmota_cmd(address, 'cycle')

        # --// CYCLE ESPHOME PORT \\--
        elif pwr_type == 'esphome':
            return self.do_esphome_cmd(address, port, 'cycle')

    def pwr_rename(self, type, address, name=None, port=None):
        if name is None:
            try:
                name = input('New name for {} port: {} >> '.format(address, port))
            except KeyboardInterrupt:
                print('Rename Aborted!')
                return 'Rename Aborted'
        if type.lower() == 'dli':
            if port is not None:
                response = self._dli[address].rename(port, name)
                if response:
                    self.data['dli_power'][address][port]['name'] = name
            else:
                response = 'ERROR port must be provided for outlet type dli'
        elif type.lower() in ['gpio', 'tasmota', 'esphome']:
            print('rename of GPIO, tasmota, and espHome ports not yet implemented')
            print('They can be renamed manually by updating ConsolePi.yaml')
            response = 'rename of GPIO, tasmota, and espHome ports not yet implemented'
            # TODO get group name based on address, read json file into dict, change the name write it back
            #      and update dict
        else:
            raise ConsolePiPowerException('pwr_rename: Invalid type ({}) or no name provided'.format(type))

        return response

    def pwr_all(self, outlets=None, action='toggle', desired_state=None):
        '''
        Returns List of responses representing state of outlet after exec
            Valid response is Bool where True = ON
            Errors are returned in str format
        '''
        if action == 'toggle' and desired_state is None:
            return 'Error: desired final state must be provided'  # should never hit this

        if outlets is None:
            outlets = self.pwr_get_outlets()['defined']
        responses = []
        for grp in outlets:
            outlet = outlets[grp]
            # if no_all: true in config outlet is ignored during all off/on operations
            if outlet.get("no_all"):
                continue
            noff = True if 'noff' not in outlet else outlet['noff']
            if action == 'toggle':
                # skip any defined dlis that don't have any linked_outlets defined
                # if not outlet['type'] == 'dli' or outlet.get('linked_devs')):
                if outlet['type'] == 'dli':
                    if outlet.get('linked_devs'):
                        responses.append(
                            self.pwr_toggle(outlet['type'], outlet['address'], desired_state=desired_state, port=self.update_linked_devs(outlet)[1], noff=noff, noconfirm=True)
                        )
                elif outlet['type'] == 'esphome':
                    _relays = utils.listify(outlet.get('relays'))
                    for p in _relays:
                        responses.append(self.pwr_toggle(outlet['type'], outlet['address'], desired_state=desired_state,
                                         port=p, noff=noff, noconfirm=True))
                else:
                    responses.append(self.pwr_toggle(outlet['type'], outlet['address'], desired_state=desired_state,
                                     noff=noff, noconfirm=True))
            elif action == 'cycle':
                if outlet['type'] == 'dli':
                    if 'linked_ports' in outlet:
                        linked_ports = utils.listify(outlet['linked_ports'])
                        for p in linked_ports:
                            # Start a thread for each port run in parallel
                            # menu status for (linked) power menu is updated on load
                            threading.Thread(
                                target=self.pwr_cycle,
                                args=[outlet['type'], outlet['address']],
                                kwargs={'port': p, 'noff': noff},
                                name=f'cycle_{p}'
                            ).start()
                elif outlet['type'] == 'esphome':
                    relays = utils.listify(outlet.get('relays', []))
                    for p in relays:
                        # Start a thread for each port run in parallel
                        threading.Thread(
                            target=self.pwr_cycle,
                            args=[outlet['type'], outlet['address']],
                            kwargs={'port': p, 'noff': noff},
                            name=f'cycle_{p}'
                        ).start()
                else:
                    threading.Thread(
                        target=self.pwr_cycle,
                        args=[outlet['type'], outlet['address']],
                        kwargs={'noff': noff},
                        name='cycle_{}'.format(outlet['address'])
                    ).start()

        # Wait for all threads to complete
        while True:
            threads = 0
            for t in threading.enumerate():
                if 'cycle' in t.name or 'toggle_' in t.name:
                    threads += 1
            if threads == 0:
                break

        return responses



#!/etc/ConsolePi/venv/bin/python3

'''
argv1 is the process menu currently uses picocom
argv2 is the device
'''
# imports in try/except in case user aborts after ssh session established
try:
    import psutil
    import sys
    import subprocess
    import time
    sys.path.insert(0, '/etc/ConsolePi/src/pypkg')
    from consolepi import config, utils  # type: ignore # NoQA
    from consolepi.consolepi import ConsolePi  # type: ignore # NoQA
    cpi = ConsolePi(bypass_remotes=True)
except (KeyboardInterrupt, EOFError):
    print("Operation Aborted")
    exit(0)


def find_procs_by_name(name, dev):
    "Return a list of processes matching 'name'."
    ppid = None
    for p in psutil.process_iter(attrs=["name", "cmdline"]):
        if name == p.info['name'] and dev in p.info['cmdline']:
            ppid = p.pid if p.ppid() == 1 else p.ppid()
    return ppid


def terminate_process(pid):
    p = psutil.Process(pid)
    x = 0
    while x < 2:
        p.terminate()
        if p.status() != 'Terminated':
            p.kill()
        else:
            break
        x += 1


def check_hung_process(cmd: str, device: str) -> int:
    ppid = find_procs_by_name(cmd, device)
    retry = 0
    msg = f"\n{_device.replace('/dev/', '')} appears to be in use (may be a previous hung session)."
    msg += '\nDo you want to Terminate the existing session'
    if ppid is not None and utils.user_input_bool(msg):
        while ppid is not None and retry < 3:
            print('Terminating Existing Session...')
            try:
                terminate_process(ppid)
                time.sleep(3)
                ppid = find_procs_by_name(cmd, device)
            except PermissionError:
                print('This appears to be a locally established session, if you want to kill that do it yourself')
                break
            except psutil.AccessDenied:
                print('This appears to be a locally established session, session will not be terminated')
                break
            except psutil.NoSuchProcess:
                ppid = find_procs_by_name(cmd, device)
            retry += 1

    return ppid


if __name__ == '__main__':
    # Allow user to ssh to configured port which using ForceCommand and specifying only the device they want to connect to
    if len(sys.argv) == 2 and "picocom" not in sys.argv[1]:
        _device = f'/dev/{sys.argv[1].replace("/dev/", "")}'
        adapter_data = cpi.local.adapters.get(_device)
        if not adapter_data:
            print(f'{_device.replace("/dev/", "")} Not found on system... Refreshing local adapters.')
            cpi.local.build_adapter_dict(refresh=True)
            adapter_data = cpi.local.adapters.get(_device)

        if adapter_data:
            print(f'Establishing Connection to {_device.replace("/dev/", "")}...')
            _cmd = adapter_data["config"]["cmd"].replace("{{timestamp}}", time.strftime("%F_%H.%M")).split()
        else:
            print(f'{_device.replace("/dev/", "")} Not found on system... Exiting.')
            sys.exit(1)
    # Sent from menu on remote full command is sent
    elif len(sys.argv) >= 3:
        _cmd = sys.argv[1:]
        _device = f'/dev/{sys.argv[2].replace("/dev/", "")}'

    ppid = check_hung_process(_cmd[0], _device)

    if ppid is None:
        # if power feature enabled and adapter linked - ensure outlet is on
        # TODO try/except block here
        if config.power:
            cpi.cpiexec.exec_auto_pwron(_device)
        subprocess.run(_cmd)



#!/etc/ConsolePi/venv/bin/python3

import pyudev
import socket
import netifaces as ni
import os
import time
from pathlib import Path
from consolepi import utils, log, config  # type: ignore


class Local():
    ''' Class to collect and manage ConsolePis local attributes '''

    def __init__(self):
        self.default_baud = config.default_baud
        self.api_port = config.api_port
        self.udev_adapters = self.detect_adapters()
        self.adapters = self.build_adapter_dict()
        self.hostname = socket.gethostname()
        self.cpuserial = self.get_cpu_serial()
        self.interfaces = self.get_if_info()
        self.ip_list = self.get_ip_list()
        self.data = self.build_local_dict()
        self.user = config.loc_user
        self.loc_home = os.path.expanduser(f'~{self.user}')

    def build_local_dict(self, rem_ip=None, refresh=False):
        '''Display representation of all local data in combined dict.'''
        if refresh:
            self.adapters = self.build_adapter_dict(refresh=True)
            self.interfaces = self.get_if_info()
        _ip_w_gw = self.interfaces['_ip_w_gw']
        rem_ip = _ip_w_gw if rem_ip is None else rem_ip
        local = {
            self.hostname: {
                'cpuserial': self.cpuserial,
                'adapters': self.adapters,
                'interfaces': self.interfaces,
                'rem_ip': rem_ip,
                'api_port': self.api_port,
                'user': config.cfg.get('rem_user', 'pi')
            }
        }
        return local

    # TODO pyudev.Devices.time_since_initialized is datetime timedelta dt.now() - dev.time_since_initialized.  Include init_time as attribute of adapter object
    def detect_adapters(self, key=None):
        """Detect Locally Attached Adapters.

        Returns
        -------
        dict
            udev alias/symlink if defined/found as key or root device if not.
            /dev/ is stripped: (ttyUSB0 | AP515).  Each device has it's attrs
            in a dict.
        """
        if key is not None:
            key = '/dev/' + key.split('/')[-1]  # key can be provided with or without /dev/ prefix

        context = pyudev.Context()

        devs = {'_dup_ser': {}}
        usb_list = [dev.properties['DEVPATH'].split('/')[-1] for dev in context.list_devices(ID_BUS='usb', subsystem='tty')]
        pci_list = [dev.properties['DEVPATH'].split('/')[-1] for dev in context.list_devices(ID_BUS='pci', subsystem='tty')]
        i2c_list = [dev.properties['DEVPATH'].split('/')[-1] for dev in context.list_devices(ID_BUS='i2c', subsystem='tty')]
        ama_list = [dev.replace('/dev/', '') for dev in config.cfg_yml.get('TTYAMA', {})]
        root_dev_list = usb_list + pci_list + i2c_list + ama_list

        for root_dev in root_dev_list:
            # determine if the device already has a udev alias & collect available path options for use on lame adapters
            dev_name = by_path = by_id = None
            try:
                _dev = pyudev.Devices.from_name(context, 'tty', root_dev)
            except pyudev._errors.DeviceNotFoundByNameError:
                log.error(f'pyudev Ubable to find {root_dev}')
                continue  # TODO Catching error as have seen it in consolepi-mdnsreg not sure if continue is appropriate
            _devlinks = _dev.get('DEVLINKS', '').split()
            if not _devlinks:   # skip occurs on non rpi and ttyAMA
                if not root_dev.startswith('ttyAMA'):
                    continue
            else:
                for _d in _devlinks:
                    if '/dev/serial' not in _d:
                        dev_name = _d.replace('/dev/', '')
                    elif '/dev/serial/by-path/' in _d:
                        by_path = _d
                    elif '/dev/serial/by-id/' in _d:
                        by_id = _d

            dev_name = f'/dev/{root_dev}' if not dev_name else f'/dev/{dev_name}'
            devs[dev_name] = {'by_path': by_path, 'by_id': by_id}
            devs[dev_name]['root_dev'] = True if dev_name == f'/dev/{root_dev}' else False

            # Gather all available properties from device  # TODO investigate use of pyudev.Device.time_since_initialized
            _props = {p.lower() if p != 'ID_USB_INTERFACE_NUM' else 'id_ifnum': _dev.properties[p]
                      for p in _dev.properties}
            devs[dev_name] = {**devs[dev_name], **_props}

            # -- no need for remaining logic on ttyAMA adapters (local UART)
            if 'ttyAMA' in root_dev:
                # TODO clean up logic this is copy paste from below
                # clean up some redundant or less useful properties
                rm_list = ['devlinks', 'id_mm_candidate', 'id_model_enc', 'id_path_tag', 'tags', 'major', 'minor',
                           'usec_initialized', 'id_vendor_enc', 'id_pci_interface_from_database', 'id_revision']

                devs[dev_name] = {k: v for k, v in devs[dev_name].items() if k not in rm_list}

                continue

            # with some multi-port adapters the model_id and vendor_id need to be pulled from higher in stack
            this_dev = _dev
            while '0x' in this_dev.properties.get('ID_MODEL_ID', '0x') and hasattr(this_dev, 'parent'):
                this_dev = this_dev.parent

            # -- Collect path for mapping to specific USB port
            # TODO clean this up not efficient could combine search for ID_MODEL_ID and devpath
            lame_devpath = this_dev.attributes.get('devpath')
            if lame_devpath and isinstance(lame_devpath, bytes):
                lame_devpath = lame_devpath.decode('UTF-8')
            else:
                for p in _dev.ancestors:
                    if 'devpath' in p.attributes.available_attributes:
                        lame_devpath = p.attributes.get('devpath')
                        if lame_devpath and isinstance(lame_devpath, bytes):
                            lame_devpath = lame_devpath.decode('UTF-8')
                            break

            devs[dev_name]['lame_devpath'] = lame_devpath

            fallback_ser = this_dev.properties.get('ID_SERIAL_SHORT')
            devs[dev_name]['id_model_id'] = this_dev.properties['ID_MODEL_ID']
            devs[dev_name]['id_vendor_id'] = this_dev.properties['ID_VENDOR_ID']
            devs[dev_name]['time_since_init'] = f'{_dev.properties.device.time_since_initialized} ' \
                                                f"as of {time.strftime('%x %I:%M:%S %p %Z', time.localtime(time.time()))}"

            # clean up some redundant or less useful properties
            rm_list = ['devlinks', 'id_mm_candidate', 'id_model_enc', 'id_path_tag', 'tags', 'major', 'minor',
                       'usec_initialized', 'id_vendor_enc', 'id_pci_interface_from_database', 'id_revision']

            devs[dev_name] = {k: v for k, v in devs[dev_name].items() if k not in rm_list}

            # --- // Handle Multi-Port adapters that use same serial for all interfaces \\ ---
            # Capture the dict in dup_ser it's later del if no additional devices present with the same serial
            # Capture path and ifnum for any subsequent devs if ser is already in the dup_ser dict
            _ser = devs[dev_name]['id_serial_short'] = _dev.get('ID_SERIAL_SHORT', fallback_ser)
            if _ser not in devs['_dup_ser']:
                devs['_dup_ser'][_ser] = {'id_paths': [], 'id_ifnums': []}

            devs['_dup_ser'][_ser]['id_paths'].append(devs[dev_name]['id_path'])
            devs['_dup_ser'][_ser]['id_ifnums'].append(devs[dev_name]['id_ifnum'])

        # --- // Clean up detection of Multi-Port adapters that use same serial for all interfaces \\ ---
        # all dev serial #s are added to dup_ser as they are discovered
        # remove any serial #s that only appeared once.
        del_list = [_ser for _ser in devs['_dup_ser'] if len(devs['_dup_ser'][_ser]['id_paths']) == 1]
        for i in del_list:
            del devs['_dup_ser'][i]

        return devs if key is None else devs[key]

    def default_ser_config(self, tty_dev, tty_port=0):
        '''Return default serial parameters when no match found in ser2net'''
        return {
            'port': tty_port,
            'baud': self.default_baud,
            'dbits': 8,
            'parity': 'n',
            'flow': 'n',
            'sbits': 1,
            'logfile': None,
            'cmd': f'picocom {tty_dev} --baud {self.default_baud}'
        }

    def build_adapter_dict(self, refresh=False):
        '''Create final adapter dict from udev ser2net and outlet dicts.'''
        if refresh or not hasattr(self, 'udev_adapters'):
            self.udev_adapters = self.detect_adapters()
        udev = {a: self.udev_adapters[a] for a in self.udev_adapters if a != '_dup_ser'}
        linked = [] if not config.outlets else config.outlets['linked']
        ser2net = {} if not config.ser2net_conf else config.ser2net_conf

        adapters = {}
        for a in udev:
            adapters[a] = {'udev': udev[a],
                           'outlets': [] if a not in linked else linked[a],
                           'config': self.default_ser_config(a) if a not in ser2net else ser2net[a]
                           }

        if refresh:
            self.adapters = adapters
            self.data = self.build_local_dict()

        return adapters

    def get_cpu_serial(self):
        res = utils.do_shell_cmd("/bin/cat /proc/cpuinfo | grep Serial | awk '{print $3}'", return_stdout=True)
        if res[0] > 0:
            log.warning('Unable to get unique identifier for this pi (cpuserial)', show=True)
            return 0
        elif not res[1]:
            machine_id_file = Path("/etc/machine-id")
            if machine_id_file.exists():
                return machine_id_file.read_text().strip()
            else:
                return 0

    def get_if_info(self):
        '''Build and return dict with interface info.'''
        if_list = [i for i in ni.interfaces() if i != 'lo' and 'docker' not in i]
        if_w_gw = ni.gateways()['default'].get(ni.AF_INET, {1: None})[1]
        if_data = {_if: {'ip': ni.ifaddresses(_if).get(ni.AF_INET, {0: {}})[0].get('addr'),
                         'mac': ni.ifaddresses(_if).get(ni.AF_LINK, {0: {}})[0].get('addr'),
                         'isgw': True if _if == if_w_gw else False} for _if in if_list
                   if ni.ifaddresses(_if).get(ni.AF_INET, {0: {}})[0].get('addr')
                   }

        if_data['_ip_w_gw'] = if_data.get(if_w_gw, {'ip': None})['ip']
        log.debugv('[GET IFACES] Completed Iface Data: {}'.format(if_data))
        return if_data

    def get_ip_list(self):
        return [
            ni.ifaddresses(i).get(ni.AF_INET, {0: {}})[0].get('addr')
            for i in ni.interfaces()
            if i != 'lo' and 'docker' not in i and 'ifb' not in i and ni.ifaddresses(i).get(ni.AF_INET, {0: {}})[0].get('addr')
        ]


if __name__ == '__main__':
    pass



#!/etc/ConsolePi/venv/bin/python3

# Original Author: Andreas Spiess
# Modified by Wade ~ Pack3tL0ss
#
# Provided with ConsolePi but not a required part of ConsolePi
# This script can be used to provide dynamic variable speed support
# for a fan via GPIO
#    currently the desired_temp GPIO pins and other variables are statically configured in this script
#        I will probably add an option to configure it in ConsolePi.yaml so updates to this script
#        don't revert your settings if you want to tweak them.
#          To customize, you can copy this script, then modify the unit file
#           to reference the new location
#
import logging
import logging.handlers
import os
import subprocess
from pathlib import Path
from time import sleep

import RPi.GPIO as GPIO

# Not elegant, but until I figure out how to pass an env var into venv
_fan_debug = os.getenv("FAN_DEBUG", "").lower()
DEBUG = True if _fan_debug == "true" or _fan_debug == "1" or Path("/tmp/FAN_DEBUG").is_file() else False

# -- // LOGGING \\ --
log = logging.getLogger("gpiofand")
log.setLevel(logging.INFO if not DEBUG else logging.DEBUG)
handler = logging.handlers.SysLogHandler(address="/dev/log")
log.addHandler(handler)

fan_pin = 4
battery_mon_pin = 18  # scl
DEFAULT_INTERVAL = 30  # Number of seconds between temp checks
BATTERY_MON = False  # disable / enable battery monitor auto-shutdown

desired_temp = 40  # The maximum temperature in Celsius after which we trigger the fan
# desired_temp = 31 # The maximum temperature in Celsius after which we trigger the fan

fan_speed = 0
prev_fan_speed = 0
sum = 0
pTemp = 15
iTemp = 0.4
interval = DEFAULT_INTERVAL


def do_shutdown():
    fan_off()
    os.system("sudo shutdown -h 1")
    sleep(100)


def get_temp():
    temp = subprocess.run(
        "vcgencmd measure_temp",
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=True,
    )
    temp = temp.stdout.decode("UTF-8").replace("temp=", "").replace("'C\n", "")
    log.debug(f"temp is {temp}'C    {round((float(temp) * 1.8), 1) + 32}'F")
    return temp


def fan_off(myPWM):
    myPWM.ChangeDutyCycle(0)  # switch fan off
    return


def handle_fan(myPWM):
    global fan_speed, sum, prev_fan_speed
    actual_temp = float(get_temp())
    diff = actual_temp - desired_temp
    sum = sum + diff
    pDiff = diff * pTemp
    iDiff = sum * iTemp
    fan_speed = round(pDiff + iDiff, 1)
    if fan_speed > 100:
        fan_speed = 100
    elif fan_speed < 15:
        fan_speed = 0

    if sum > 100:
        sum = 100
    elif sum < -100:
        sum = -100

    if fan_speed != prev_fan_speed:
        f_temp = round((actual_temp * 1.8) + 32, 1)
        # We increase the time between checks to 5 mins if the temp is 10+ over desired
        if not BATTERY_MON:
            interval = 320 if diff > 10 else DEFAULT_INTERVAL

        log.info(
            f"Fan Speed change: {prev_fan_speed}% --> {fan_speed}% CPU Temp: {f_temp}'F, "
            f"Desired Temp {(desired_temp * 1.8) + 32}'F, Next Check in {interval}s"  # type: ignore
        )
        prev_fan_speed = fan_speed
    log.debug(
        "actual_temp %4.2f desiredTemp %4.2f TempDiff %4.2f pDiff %4.2f iDiff %4.2f fan_speed %5d"
        % (
            round(actual_temp * 1.8 + 32, 1),
            desired_temp * 1.8 + 32,
            diff,
            pDiff,
            iDiff,
            fan_speed,
        )
    )
    myPWM.ChangeDutyCycle(fan_speed)
    return


def handle_battery():
    log.debug(GPIO.input(battery_mon_pin))
    if GPIO.input(battery_mon_pin) == 0:
        ("do_shutdown()")
        sleep(5)
        do_shutdown()
    return


# Unused debug function
def set_pin(mode):
    GPIO.output(fan_pin, mode)
    return


def main():
    log.debug(f"--- GPIO Fan Monitor Startup DEBUG Enabled ---")
    try:
        GPIO.setwarnings(False)
        GPIO.setmode(GPIO.BCM)
        GPIO.setup(fan_pin, GPIO.OUT)
        myPWM = GPIO.PWM(fan_pin, 50)
        myPWM.start(50)

        if BATTERY_MON:
            GPIO.setup(battery_mon_pin, GPIO.IN, pull_up_down=GPIO.PUD_DOWN)
        fan_off(myPWM)
        while True:
            handle_fan(myPWM)
            if BATTERY_MON:
                handle_battery()
            sleep(interval)
    except KeyboardInterrupt:
        fan_off(myPWM)  # type: ignore
        GPIO.cleanup()  # resets all GPIO ports used by this program
    log.info(f"--- GPIO Fan Monitor ShutDown Fan is off---")


if __name__ == "__main__":
    main()



#!/etc/ConsolePi/venv/bin/python3

import json
import os
import os.path
import pickle
import socket
import sys
import time

# -- google stuff --
from google.auth.transport.requests import Request
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient import discovery

sys.path.insert(0, '/etc/ConsolePi/src/pypkg')
from consolepi import config, log, utils  # type: ignore # NoQA


# -- GLOBALS --
TIMEOUT = 3   # socket timeout in seconds (for clustered/cloud setup)
SCOPES = ['https://www.googleapis.com/auth/spreadsheets', 'https://www.googleapis.com/auth/drive.metadata']


class GoogleDrive:

    def __init__(self, hostname=None):
        self.hostname = socket.gethostname() if hostname is None else hostname
        self.creds = None
        self.file_id = None
        self.sheets_svc = None

    def exec_request(self, _request):
        result = None
        attempt = 0
        while True:
            attempt += 1
            try:
                result = _request.execute()
                break
            except Exception as e:
                log.show(f'Exception while communicating with Gdrive: {e.__class__.__name__}', show=True)
                log.exception(f'[GDRIVE]: Exception while communicating with Gdrive\n{e}')
            if attempt > 2:
                log.error(('[GDRIVE]: Giving up after {} attempts'.format(attempt)))
                break
        return result

    # Authenticate find file_id and build services
    def auth(self):
        if utils.is_reachable('www.googleapis.com', 443):
            try:
                if self.creds is None:
                    self.creds = self.get_credentials()
                if self.sheets_svc is None:
                    self.sheets_svc = discovery.build('sheets', 'v4', credentials=self.creds, cache_discovery=False)
                if self.file_id is None:
                    self.file_id = self.get_file_id()
                    if self.file_id is None:
                        self.file_id = self.create_sheet()
                return True
            except (ConnectionError, TimeoutError, OSError) as e:
                log.error(f'Exception Occurred Connecting to Gdrive {e.__class__.__name__}')
                return False
        else:
            log.error('Google Drive is not reachable - Aborting')
            return False

    # Google sheets API credentials - used to update config on Google Drive
    def get_credentials(self):
        '''Get credentials for google drive / sheets

        Returns:
            Object -- Credentials Object
        '''
        log.debug('[GDRIVE]: -->get_credentials() {}'.format(log.name))
        creds = None
        # The file token.pickle stores the user's access and refresh tokens, and is
        # created automatically when the authorization flow completes for the first
        # time.
        os.chdir('/etc/ConsolePi/cloud/gdrive/')
        if os.path.exists('.credentials/token.pickle'):
            with open('.credentials/token.pickle', 'rb') as token:
                creds = pickle.load(token)
        # If there are no (valid) credentials available, let the user log in.
        if not creds or not creds.valid:
            if creds and creds.expired and creds.refresh_token:
                creds.refresh(Request())
            else:
                flow = InstalledAppFlow.from_client_secrets_file(
                    '.credentials/credentials.json', SCOPES)
                creds = flow.run_local_server()
            # Save the credentials for the next run
            with open('.credentials/token.pickle', 'wb') as token:
                pickle.dump(creds, token)
        return creds

    # Google Drive Get File ID from File Name
    def get_file_id(self):
        '''Gets file id associated with ConsolePi.csv on Google Drive.

        Returns:
            str -- file id for ConsolePi.csv if cound otherwise No return
        '''
        if self.creds is None:
            self.creds = self.get_credentials()

        service = discovery.build('drive', 'v3', credentials=self.creds, cache_discovery=False)
        results = service.files().list(
            pageSize=10, fields="nextPageToken, files(id, name)", q="name='ConsolePi.csv'").execute()
        items = results.get('files', [])
        if not items:
            print('No files found.')
        else:
            for item in items:
                if item['name'] == 'ConsolePi.csv':
                    return item['id']

    # Create spreadsheet to store data if not found (get_file_id returns None)
    def create_sheet(self):
        service = self.sheets_svc
        log.info('[GDRIVE]: ConsolePi.csv not found on Gdrive. Creating ConsolePi.csv')
        spreadsheet = {
            'properties': {
                'title': 'ConsolePi.csv'
            }
        }
        request = service.spreadsheets().create(body=spreadsheet,
                                                fields='spreadsheetId')
        spreadsheet = self.exec_request(request)
        return '{0}'.format(spreadsheet.get('spreadsheetId'))

    # Auto Resize gdrive columns to match content
    def resize_cols(self):
        service = self.sheets_svc
        body = {"requests": [{"autoResizeDimensions": {"dimensions": {"sheetId": 0, "dimension": "COLUMNS",
                                                                      "startIndex": 0, "endIndex": 2}}}]}
        request = service.spreadsheets().batchUpdate(
            spreadsheetId=self.file_id,
            body=body)
        response = self.exec_request(request)
        log.debug('[GDRIVE]: resize_cols response: {}'.format(response))

    def update_files(self, data):
        for x in data[self.hostname]['adapters']:
            if 'udev' in data[self.hostname]['adapters'][x]:
                del data[self.hostname]['adapters'][x]['udev']

        log.debugv('[GDRIVE]: -->update_files - data passed to function\n{}'.format(json.dumps(data, indent=4, sort_keys=True)))
        if not self.auth():
            return 'Gdrive-Error: Unable to Connect to Gdrive refer to cloud log for details'
        spreadsheet_id = self.file_id
        service = self.sheets_svc

        # init remote_consoles dict, any entries in config not matching this ConsolePis hostname are added as remote ConsolePis
        value_input_option = 'USER_ENTERED'
        remote_consoles = {}
        cnt = 1
        data[self.hostname]['upd_time'] = int(time.time())  # Put timestamp (epoch) on data for this ConsolePi
        for k in data:
            found = False
            value_range_body = {
                "values": [
                    [
                        k,
                        json.dumps(data[k])
                    ]
                ]
            }

            # find out if this ConsolePi already has a row use that row in range
            request = service.spreadsheets().values().get(
                spreadsheetId=spreadsheet_id, range='A:B')
            result = self.exec_request(request)
            log.info('[GDRIVE]: Reading from Cloud Config')
            if result and result.get('values') is not None:
                x = 1
                for row in result.get('values'):
                    if k == row[0]:  # k is hostname row[0] is column A of current row
                        cnt = x
                        found = True
                    else:
                        log.info('[GDRIVE]: {0} found in Google Drive Config'.format(row[0]))
                        remote_consoles[row[0]] = json.loads(row[1])
                        remote_consoles[row[0]]['source'] = 'cloud'
                    x += 1
                log.debugv(f'[GDRIVE]: {len(remote_consoles)} Remote ConsolePis Found on Gdrive: \n{json.dumps(remote_consoles)}')
            range_ = 'a' + str(cnt) + ':b' + str(cnt)

            # -- // Update gdrive with this ConsolePis data \\ --
            if not config.cloud_pull_only:
                if found:
                    log.info('[GDRIVE]: Updating ' + str(k) + ' data found on row ' + str(cnt) + ' of Google Drive config')
                    request = service.spreadsheets().values().update(spreadsheetId=spreadsheet_id, range=range_,
                                                                     valueInputOption=value_input_option, body=value_range_body)
                else:
                    log.info('[GDRIVE]: Adding ' + str(k) + ' to Google Drive Config')
                    request = service.spreadsheets().values().append(spreadsheetId=spreadsheet_id, range=range_,
                                                                     valueInputOption=value_input_option, body=value_range_body)
                self.exec_request(request)
            else:
                log.info('cloud_pull_only override enabled not updating cloud with data from this host')
            cnt += 1
        self.resize_cols()
        return remote_consoles


if __name__ == '__main__':
    print('-- Syncing Data With Google Drive --')
    from consolepi.local import Local  # type: ignore
    local = Local()
    gdrive = GoogleDrive()
    data = gdrive.update_files(local.data)
    print(json.dumps(data, indent=4, sort_keys=True) if 'Gdrive-Error' not in data else data)



#!/etc/ConsolePi/venv/bin/python3

import sys
from pathlib import Path
import shutil
from rich import print
from rich.prompt import Confirm
from datetime import datetime
from typing import List

sys.path.insert(0, '/etc/ConsolePi/src/pypkg')
from consolepi import config  # type: ignore # NoQA

DEF_V4_FILE = Path("/etc/ConsolePi/src/ser2net.yaml")
V3_FILE = Path("/etc/ser2net.conf")
BAK_DIR = Path("/etc/ConsolePi/bak")
DEF_PROMPT = f"""
[green]--------------------[/] [cyan]ser2net[/] [bright_red]v3[/] to [green3]v4[/] conversion utility [green]--------------------[/]
This utility will parse the [cyan]ser2net[/] v3 configuration file [cyan italic]{V3_FILE.name}[/]
and create a ConsolePi compatible ser2net v4 config [cyan italic]ser2net.yaml[/].

If ser2net v4 is installed the v3 config will be moved to {BAK_DIR}.
"""


class Convert:
    def __init__(self):
        self.v4_file: Path = self.get_v4_config_name()

    def __call__(self):
        if "-y" in " ".join(sys.argv[1:]).lower() or self.confirm():
            _ = self.check_v4_config()
            v4_lines = self.gather_config_data()
            if self.v4_file.exists() and self.v4_file.read_text() == v4_lines:
                print(f'  [magenta]-[/] Process resulted in no change in {self.v4_file.name} content.  Existing v4 config left as is.')
            else:
                print(f'  [magenta]-[/] Writing converted config to {self.v4_file}')
                self.v4_file.write_text(v4_lines)
                self.backup_v3_config()
            print()
        else:
            print("[bright_red]Aborted.")

    def get_v4_config_name(self) -> Path:
        new_file = Path("/etc/ser2net.yaml")
        if not new_file.exists() and Path("/etc/ser2net.yml").exists():
            new_file = Path("/etc/ser2net.yml")

        return new_file

    def check_v4_config(self) -> bool:
        if self.v4_file.exists():
            if "ConsolePi" not in self.v4_file.read_text().splitlines()[2]:
                now = datetime.now()
                _bak_file = BAK_DIR / f'{self.v4_file.name}.{now.strftime("%F_%H%M")}'
                print(f'  [magenta]-[/] Backing up existing non ConsolePi version of [cyan]{self.v4_file}[/] to [cyan]{_bak_file}[/].')
                shutil.move(self.v4_file, _bak_file)
                do_ser2net = True
                print("  [magenta]-[/] Preparing default ConsolePi ser2net v4 base config.")
            else:
                print(f'  [magenta]-[/] Found existing ConsolePi compatible version of [cyan]{self.v4_file}[reset].')
                do_ser2net = False
        else:
            do_ser2net = True
            print("  [magenta]-[/] Preparing default ConsolePi ser2net v4 base config.")

        return do_ser2net

    def backup_v3_config(self):
        if int("".join(config.ser2net_ver.split(".")[0])) > 3:
            if not self.v4_file.exists:
                print(f'  [dark_orange3][blink]!![/blink] WARNING[/]: ser2net v4 Configuration ({self.v4_file}) not found.  Keeping ser2net v3 file ({V3_FILE}) in place')
            else:
                now = datetime.now()
                _bak_file = BAK_DIR / f'{V3_FILE.name}.{now.strftime("%F_%H%M")}'
                print(f"  [magenta]-[/] ser2net [cyan]v{config.ser2net_ver}[/] installed.  Backing up v3 config {V3_FILE.name} to {_bak_file}")
                shutil.move(V3_FILE, _bak_file)
        else:
            print(f"  [magenta]-[/] ser2net [cyan]v{config.ser2net_ver}[/] installed.  Keeping v3 config {V3_FILE} in place.")

    def gather_config_data(self) -> str:
        v3_aliases = {k: v for k, v in config.ser2net_conf.items() if not k.startswith("_") and 7000 < v.get('port', 0) <= 7999}
        if self.v4_file.exists():
            v4_config: List[str] = self.v4_file.read_text().splitlines(keepends=True)
            v4_config_dict = config.get_ser2netv4(self.v4_file)
            v4_aliases = {k: v for k, v in v4_config_dict.items() if not k.startswith("_") and 7000 < v.get('port', 0) <= 7999}
        else:
            v4_config: List[str] = DEF_V4_FILE.read_text().splitlines(keepends=True)
            v4_aliases = {}

        v4_user_lines = [v["v4line"] for k, v in v3_aliases.items() if k not in v4_aliases.keys()]

        # Gather any trace-file references from v3 config and ensure they are defined in v4 config.
        if "trace-" in "".join(v4_user_lines):
            trace_lines = [line.strip().split(":")[-1].strip() for line in v4_user_lines if "trace-" in line]
            new_defs = [ref.lstrip("*") for ref in trace_lines if ref.startswith("*")]  # trace-file defs referenced in old config
            if new_defs:
                existing_defs, get_next_empty, insert_idx = [], False, 19
                for idx, line in enumerate(v4_config):
                    if "define: &" in line and "banner" not in line and "/" in line:
                        existing_defs += [line.split("&")[-1].split()[0]]  # define: &trace3 /var/log/ser2net/trace3...
                    elif get_next_empty:
                        if not line.strip() or not line.strip().startswith("define:"):
                            get_next_empty = False
                            insert_idx = idx
                    elif "# TRACEFILE DEFINITIONS" in line.strip():
                        get_next_empty = True
                    else:
                        continue

                missing_defs = [ref for ref in new_defs if ref not in existing_defs]
                if missing_defs:
                    missing_lines = [line for line in config.ser2net_conf.get("_v4_tracefiles", "").splitlines() for d in missing_defs if not line.startswith("#") and f'define: &{d}' in line]
                    if missing_lines:
                        print(f"  [magenta]-[/] Adding trace-file definition{'s' if len(missing_defs) > 1 else ''} [cyan]{'[/], [cyan]'.join(missing_defs)}[/] referenced in v3 config.")
                        v4_config.insert(insert_idx, "\n".join(missing_lines) + "\n")

        if v4_user_lines:
            new_cnt = len([line for line in v4_user_lines if line.startswith("connection:")])
            print(f"  [magenta]-[/] Adding {new_cnt} alias definitions converted from {config.ser2net_file}")
            v4_config += v4_user_lines
            if len(v3_aliases) != new_cnt:
                print(f"  [magenta]-[/] {len(v3_aliases) - new_cnt} aliases found in ser2net v3 config were skipped as they are already defined in the existing v4 config.")
        elif v3_aliases and v4_aliases:
            print(f'  [dark_orange3][blink]!![/blink] WARNING[/]: All {len(v3_aliases)} alias definitions found in the ser2net v3 config were ignored as they are already defined in the existing v4 config.')
        else:
            print(f'  [dark_orange3][blink]!![/blink] WARNING[/]: No User defined aliases found in {V3_FILE.name}.')

        return "".join(v4_config)

    @staticmethod
    def confirm(prompt: str = None) -> bool:
        prompt = prompt or DEF_PROMPT
        print(DEF_PROMPT)
        choice = Confirm.ask("Proceed?")
        print(f"[green]{'':-^77}[/]\n")

        return choice


if __name__ == "__main__":
    if shutil.os.getuid() != 0:
        print("[bright_red]Error:[/] Must be run as root.  Launch with [cyan]consolepi-convert[/].")
        sys.exit(1)
    elif not Path("/etc/ser2net.conf").exists():
        print("[bright_red]Error:[/] No v3 config found ([cyan]/etc/ser2net.conf[/]).")
        sys.exit(1)
    else:
        convert = Convert()
        convert()



#!/etc/ConsolePi/venv/bin/python3

import threading
import time
import os
import json
import subprocess
import shlex
from halo import Halo

from consolepi import utils, log, config

# TODO byobu to menu launch new sessions in new tab by default figure out best way to provide split options i.e. 11 split h 14 create new-window and launch 11 on top 14 on bottom
# Command to launch menu in byobu: byobu new-session -n menu consolepi-menu
# command to launch new window
# byobu new-window -n r1-6100-oobm "bash -c 'echo -e $pre_msg \n;stty size; ssh -t garagepi /etc/ConsolePi/src/remote_launcher.py picocom /dev/r1-6100-oobm-sw -b 115200'"
# or byobu new-window -n name 'ssh -t garagepi /etc/ConsolePi/src/remote_launcher.py picocom /dev/r1-6100-oobm-sw -b 115200'
# # temp byobu launch testing
# menu_actions['by'] = {
#     "cmd": "sudo -u wade byobu new-window -n r1-6100-oobm-sw 'sudo -u wade ssh -t garagepi /etc/ConsolePi/src/remote_launcher.py picocom /dev/r1-6100-oobm-sw -b 115200'",
#     "pre_msg": "Connecting To r1-6100-oobm-sw on GaragePi..."

class ConsolePiExec:
    def __init__(self, config, pwr, local, menu):
        self.pwr = pwr
        self.local = local
        self.menu = menu
        self.pwr_init_complete = False
        self.autopwr_wait = False
        self.spin = Halo(spinner="dots")

    def exec_auto_pwron(self, pwr_key):
        """Launch auto_pwron in thread

        params:
            menu_dev:str, The tty dev user is connecting to
        """
        pwr_key_pretty = pwr_key.replace("/dev/", "").replace("/host/", "")
        if (
            pwr_key in config.outlets["linked"]
        ):  # verify against config here as pwr may not be completely init
            _msg = (
                f"Ensuring {pwr_key_pretty} Linked Outlets ({' '.join(config.outlets['linked'][pwr_key])}) "
                "are Powered \033[1;32mON\033[0m"
            )
            _dots = "-" * (len(_msg) + 4)
            _msg = (
                f"\n{_dots}\n  {_msg}  \n{_dots}\n"  # TODO send to formatter in menu ... __init__
            )
            print(_msg)
            threading.Thread(
                target=self.auto_pwron_thread,
                args=(pwr_key,),
                name="auto_pwr_on_" + pwr_key_pretty,
            ).start()
            log.debug(
                "[AUTO PWRON] Active Threads: {}".format(
                    [t.name for t in threading.enumerate() if t.name != "MainThread"]
                )
            )

    # TODO just get the outlet from the dict, and pass to power module function let it determine type etc
    def auto_pwron_thread(self, pwr_key):
        """Ensure any outlets linked to device are powered on

        Called by consolepi_menu exec_menu function and remote_launcher (for sessions to remotes)
        when a connection initiated with adapter.  Powers any linked outlets associated with the
        adapter on.

        params:
            menu_dev:str, The tty device user is connecting to.
        Returns:
            No Return - Updates class attributes
        """
        if self.wait_for_threads("init"):
            return

        outlets = self.pwr.data
        if "linked" not in outlets:
            _msg = "Error linked key not found in outlet dict\nUnable to perform auto power on"
            log.show(_msg, show=True)
            return

        if not outlets["linked"].get(pwr_key):
            return

        # -- // Perform Auto Power On (if not already on) \\ --
        for o in outlets["linked"][pwr_key]:
            outlet = outlets["defined"].get(o.split(":")[0])
            if outlet:
                ports = [] if ":" not in o else json.loads(o.replace("'", '"').split(":")[1])
                _addr = outlet["address"]
            else:
                log.error(
                    f"Skipping Auto Power On {pwr_key} for {o}. Unable to pull outlet details from defined outlets.",
                    show=True,
                )
                log.debugv(f"Outlet Dict:\n{json.dumps(outlets)}")
                continue

            # -- // DLI web power switch Auto Power On \\ --
            #
            # TODO combine all ports from same pwr_key and sent to pwr_toggle once
            # TODO Update outlet if return is OK, then run refresh in the background to validate
            # TODO Add class attribute to cpi_menu ~ cpi_menu.new_data = "power", "main", etc
            #      Then in wait_for_input run loop to check for updates and re-display menu
            # TODO power_menu and dli_menu wait_for_threads auto power ... check cpiexec.autopwr_wait first
            #
            if outlet["type"].lower() == "dli":
                for p in ports:
                    log.debug(
                        f"[Auto PwrOn] Power ON {pwr_key} Linked Outlet {outlet['type']}:{_addr} p{p}"
                    )

                    if not outlet["is_on"][p][
                        "state"
                    ]:  # This is just checking what's in the dict not querying the DLI
                        r = self.pwr.pwr_toggle(outlet["type"], _addr, desired_state=True, port=p)
                        if isinstance(r, bool):
                            if r:
                                threading.Thread(
                                    target=self.outlet_update,
                                    kwargs={"refresh": True, "upd_linked": True},
                                    name="auto_pwr_refresh_dli",
                                ).start()
                                self.autopwr_wait = True
                        else:
                            log.warning(
                                f"{pwr_key} Error operating linked outlet @ {o}",
                                show=True,
                            )

            # -- // esphome Auto Power On \\ --
            elif outlet["type"].lower() == "esphome":
                for p in ports:
                    log.debug(
                        f"[Auto PwrOn] Power ON {pwr_key} Linked Outlet {outlet['type']}:{_addr} p{p}"
                    )
                    if not outlet["is_on"][p]["state"]:  # This is just checking what's in the dict
                        r = self.pwr.pwr_toggle(outlet["type"], _addr, desired_state=True, port=p)
                        if isinstance(r, bool):
                            self.pwr.data["defined"][o.split(":")[0]]["is_on"][p]["state"] = r
                        else:
                            log.show(r)
                            log.warning(
                                f"{pwr_key} Error operating linked outlet @ {o}",
                                show=True,
                            )

            # -- // GPIO & TASMOTA Auto Power On \\ --
            else:
                log.debug(f"[Auto PwrOn] Power ON {pwr_key} Linked Outlet {outlet['type']}:{_addr}")
                r = self.pwr.pwr_toggle(
                    outlet["type"],
                    _addr,
                    desired_state=True,
                    noff=outlet.get("noff", True) if outlet["type"].upper() == "GPIO" else True,
                )
                if isinstance(r, int) and r > 1:  # return is an error
                    r = False
                else:  # return is bool which is what we expect
                    if r:
                        self.pwr.data["defined"][o]["state"] = r
                        self.autopwr_wait = True
                        # self.pwr.pwr_get_outlets(upd_linked=True)
                    else:
                        # self.config.log_and_show(f"Error operating linked outlet {o}:{outlet['address']}", log=log.warning)
                        log.show(
                            f"Error operating linked outlet {o}:{outlet['address']}",
                            show=True,
                        )

    def exec_shell_cmd(self, cmd):
        """Determine if cmd is valid shell cmd and execute if so.

        Command will execute as the local user unless the user prefixed the
        cmd with sudo -u

        Arguments:
            cmd {str} -- Input provided by user

        Returns:
            True|None -- Return True if cmd was determined to be a bash cmd
        """
        s = subprocess
        c = cmd.replace("-u", "").replace("sudo", "").strip()
        p = "PATH=$PATH:/etc/ConsolePi/src/consolepi-commands && "
        # TODO if shutil.which(c.split()[0]):
        r = s.run(f"{p}which {c.split()[0]}", shell=True, stderr=s.PIPE, stdout=s.PIPE)
        if r.returncode == 0:
            try:
                if "sudo " not in cmd:
                    cmd = f'sudo -u {config.loc_user} bash -c "{p}{cmd}"'
                elif "sudo -u " not in cmd:
                    cmd = cmd.replace("sudo ", "")
                subprocess.run(cmd, shell=True)
                print("")
                input("Press Enter to Continue... ")
            except (KeyboardInterrupt, EOFError):
                log.show("Operation Aborted")
                print("")  # prevents header and prompt on same line in debug

            return True

    def wait_for_threads(self, name="init", timeout=10, thread_type="power"):
        """wait for parallel async threads to complete

        returns:
            bool: True if threads are still running indicating a timeout
                  None indicates no threads found ~ they have finished
        """
        start = time.perf_counter()
        do_log = False
        found = False
        while True:
            found = False
            for t in threading.enumerate():
                if name in t.name:
                    found = do_log = True
                    t.join(timeout - 1)

            if not found:
                if name == "init" and thread_type == "power":
                    if self.pwr and not self.pwr.data or not self.pwr.data.get("dli_power"):
                        self.pwr.dli_exists = False
                    self.pwr_init_complete = True
                if do_log:
                    log.info(
                        "[{0} {1} WAIT] {0} Threads have Completed, elapsed time: {2}".format(
                            name.strip("_").upper(),
                            thread_type.upper(),
                            round(time.perf_counter() - start, 2),
                        )
                    )
                break
            elif time.perf_counter() - start > timeout:
                log.error(
                    "[{0} {1} WAIT] Timeout Waiting for {0} Threads to Complete, elapsed time: {2}".format(
                        name.strip("_").upper(),
                        thread_type.upper(),
                        round(time.perf_counter() - start, 2),
                    ),
                    show=True,
                )
                return True

    # TODO REMOVE - Deprecated
    def launch_shell(self):
        iam = config.loc_user
        os.system(
            'sudo -u {0} echo PS1=\\"\\\033[1\;36mconsolepi-menu\\\033[0m:\\\w\\\$ \\" >/tmp/prompt && '  # NoQA
            'echo alias consolepi-menu=\\"exit\\" >>/tmp/prompt &&'
            "echo PATH=$PATH:/etc/ConsolePi/src/consolepi-commands >>/tmp/prompt && "
            'alias consolepi-menu=\\"exit\\" >>/tmp/prompt && '
            "echo \"launching local shell, 'exit' to return to menu\" &&"
            "sudo -u {0} bash -rcfile /tmp/prompt ; rm /tmp/prompt".format(iam)
        )

    def outlet_update(self, upd_linked=False, refresh=False, key="defined", outlets=None):
        """
        Called by consolepi-menu refresh and exec_auto_pwron (to update outlets in menu)
        """
        pwr = self.pwr
        if config.power:
            outlets = pwr.data if outlets is None else outlets
            if not self.pwr_init_complete or refresh:
                _outlets = pwr.pwr_get_outlets(
                    outlet_data=outlets.get("defined", {}),
                    upd_linked=upd_linked,
                    failures=outlets.get("failures", {}),
                )
                pwr.data = _outlets
            else:
                _outlets = outlets

            if key in _outlets:
                return _outlets[key]
            else:
                msg = f'Invalid key ({key}) passed to outlet_update. Returning "defined"'
                log.error(msg, show=True)
                return _outlets["defined"]

    def gen_copy_key(self, rem_data=None):
        """Generate public ssh key and distribute to remote ConsolePis

        Keyword Arguments:
            rem_data {tuple or list of tuples} -- each tuple should have 3 items
            0: hostname of remote, 1: rem_ip, 3: rem_user    (default: {None})

        Returns:
            {list} -- list of any errors reported, could be informational
        """
        hostname = self.local.hostname
        loc_user = self.local.user
        loc_home = self.local.loc_home

        # -- generate local key file if it doesn't exist
        # TODO pathlib
        if not os.path.isfile(loc_home + "/.ssh/id_rsa"):
            print("\nNo Local ssh cert found, generating...\n")
            utils.do_shell_cmd(
                f'sudo -u {loc_user} ssh-keygen -m pem -t rsa -C "{loc_user}@{hostname}"',
                timeout=360,
            )

        # -- copy keys to remote(s)
        if not isinstance(rem_data, list):
            rem_data = [rem_data]
        return_list = []
        for _rem in rem_data:
            rem, rem_ip, rem_user = _rem
            print(
                self.menu.format_line(
                    "{{magenta}}Attempting to copy ssh cert to " + rem + "{{norm}}"
                ).text
            )
            if not utils.is_reachable(rem_ip, 22, timeout=5, silent=True):
                return_list.append(f"{rem}: is not reachable... Skipped")
            else:
                ret = utils.do_shell_cmd(
                    f"sudo -u {loc_user} ssh-copy-id {rem_user}@{rem_ip}", timeout=60  # 360
                )
                if ret is not None:
                    return_list.append("{}: {}".format(rem, ret))
        return return_list

    def show_adapter_details(self, adapters):
        for a in adapters:
            print(f' --- Details For {a.replace("/dev/", "")} --- ')
            for k in sorted(adapters[a]["udev"].keys()):
                print(f'{k}: {adapters[a]["udev"][k]}')
            print("")
            # this_ser2net = config.ser2net_conf.get(a, {})
            # print(f"ser2net config: {this_ser2net.get('line', '!! Not Found !!')}")
            print(f'ser2net config: {adapters[a]["config"].get("line", "Not Defined")}')

        input("\nPress Enter To Continue\n")

    # ------ // EXECUTE MENU SELECTIONS \\ ------ #
    def menu_exec(self, choice, menu_actions, calling_menu="main_menu"):
        """Execute Menu Selection.  This method needs to be overhauled.

        The ConsolePiAction object defined but not used in __init__ is part of the plan for overhaul
        menu will build instance of the object for each selection. That will be used to determine what
        action to perform and what to do after etc.
        """
        pwr = self.pwr

        if not config.debug and calling_menu not in ["dli_menu", "power_menu"]:
            os.system("clear")

        if not choice.lower or choice.lower in menu_actions and menu_actions[choice.lower] is None:
            # They just hit return with no input
            (
                self.menu.rows,
                self.menu.cols,
            ) = utils.get_tty_size()  # re-calc tty size in case they've adjusted the window
            return

        else:
            ch = choice.lower
            try:  # Invalid Selection, Keyboard Interupt
                if isinstance(menu_actions[ch], dict):
                    if menu_actions[ch].get("cmd"):
                        # TimeStamp for picocom session log file if defined
                        menu_actions[ch]["cmd"] = menu_actions[ch]["cmd"].replace(
                            "{{timestamp}}", time.strftime("%F_%H.%M")
                        )

                        # -- // AUTO POWER ON LINKED OUTLETS \\ --
                        if config.power and "pwr_key" in menu_actions[ch]:
                            self.exec_auto_pwron(menu_actions[ch]["pwr_key"])

                        # -- // Print pre-connect messsage if provided \\ --
                        if menu_actions[ch].get("pre_msg"):
                            print(menu_actions[ch]["pre_msg"])
                            c = menu_actions[ch].get("cmd")
                            if (
                                calling_menu != "rename_menu"
                                and "/dev/" in c
                                or "telnet" in c.lower()
                            ):
                                print(self.menu.tty)

                        # --// execute the command \\--
                        try:
                            _error = None
                            if "exec_kwargs" in menu_actions[ch]:
                                c = menu_actions[ch]["cmd"]
                                _error = utils.do_shell_cmd(c, **menu_actions[ch]["exec_kwargs"])
                                if _error and self.autopwr_wait:
                                    # TODO simplify this after moving to action object
                                    _h = None
                                    _method = "ssh -t" if "ssh" in c else "telnet"
                                    if "ssh" in _method:
                                        _h = c.split(f"{_method} ")[1].split(" -p")[0].split("@")[1]
                                        _p = int(c.split("-p ")[1])
                                    elif _method == "telnet":
                                        c_list = c.split()
                                        _h = c_list[-2]
                                        _p = int(c_list[-1])

                                    def wait_for_boot():
                                        while True:
                                            try:
                                                if utils.is_reachable(_h, _p, silent=True):
                                                    break
                                                else:
                                                    time.sleep(3)
                                            except (KeyboardInterrupt, EOFError):
                                                self.autopwr_wait = False
                                                log.show("Connection Aborted")
                                                break

                                    if _h:
                                        print(
                                            "\nInitial Attempt Failed, but host is linked to an outlet that was"
                                        )
                                        print("off. Host may still be booting\n")
                                        utils.spinner(
                                            f"Waiting for {_h} to boot, CTRL-C to Abort",
                                            wait_for_boot,
                                        )
                                        if self.autopwr_wait:
                                            _error = utils.do_shell_cmd(
                                                c, **menu_actions[ch]["exec_kwargs"]
                                            )
                                            self.autopwr_wait = False
                            else:
                                c = shlex.split(menu_actions[ch]["cmd"])
                                result = subprocess.run(c, stderr=subprocess.PIPE)
                                _stderr = result.stderr.decode("UTF-8")
                                if _stderr or result.returncode == 1:
                                    _error = utils.error_handler(c, _stderr)

                            if _error:
                                log.show(_error)

                            # -- // resize the terminal to handle serial connections that jack the terminal size \\ --
                            if (
                                "picocom" in menu_actions[ch]["cmd"]
                                or "telnet" in menu_actions[ch]["cmd"]
                            ):
                                os.system("/etc/ConsolePi/src/consolepi-commands/resize >/dev/null")

                        except (KeyboardInterrupt, EOFError):
                            log.show("Aborted last command based on user input")

                    elif "function" in menu_actions[ch]:
                        args = menu_actions[ch]["args"] if "args" in menu_actions[ch] else []
                        kwargs = menu_actions[ch]["kwargs"] if "kwargs" in menu_actions[ch] else {}
                        confirmed, spin_text, name = self.confirm_and_spin(
                            menu_actions[ch], *args, **kwargs
                        )
                        if confirmed:
                            # update kwargs with name from confirm_and_spin method
                            if menu_actions[ch]["function"].__name__ == "pwr_rename":
                                kwargs["name"] = name

                            # // -- CALL THE FUNCTION \\--
                            if spin_text:  # start spinner if spin_text set by confirm_and_spin
                                with Halo(text=spin_text, spinner="dots2"):
                                    response = menu_actions[ch]["function"](*args, **kwargs)
                            else:  # no spinner
                                response = menu_actions[ch]["function"](*args, **kwargs)

                            # --// Power Menus \\--
                            if calling_menu in ["power_menu", "dli_menu"]:
                                if menu_actions[ch]["function"].__name__ == "pwr_all":
                                    with Halo(text="Refreshing Outlet States", spinner="dots"):
                                        self.outlet_update(
                                            refresh=True, upd_linked=True
                                        )  # TODO can I move this to Outlets Class
                                else:
                                    _grp = menu_actions[ch]["key"]
                                    _type = menu_actions[ch]["args"][0]
                                    _addr = menu_actions[ch]["args"][1]
                                    # --// EVAL responses for dli outlets \\--
                                    if _type == "dli":
                                        host_short = utils.get_host_short(_addr)
                                        _port = menu_actions[ch]["kwargs"]["port"]
                                        # --// Operations performed on ALL outlets \\--
                                        if isinstance(response, bool) and _port is not None:
                                            if (
                                                menu_actions[ch]["function"].__name__
                                                == "pwr_toggle"
                                            ):
                                                self.spin.start(
                                                    "Request Sent, Refreshing Outlet States"
                                                )
                                                # threading.Thread(target=self.get_dli_outlets,
                                                # kwargs={'upd_linked': True, 'refresh': True}, name='pwr_toggle_refresh').start()
                                                upd_linked = (
                                                    True if calling_menu == "power_menu" else False
                                                )  # else dli_menu
                                                threading.Thread(
                                                    target=self.outlet_update,
                                                    kwargs={
                                                        "upd_linked": upd_linked,
                                                        "refresh": True,
                                                    },
                                                    name="pwr_toggle_refresh",
                                                ).start()
                                                if _grp in pwr.data["defined"]:
                                                    pwr.data["defined"][_grp]["is_on"][_port][
                                                        "state"
                                                    ] = response
                                                elif _port != "all":
                                                    pwr.data["dli_power"][_addr][_port][
                                                        "state"
                                                    ] = response
                                                else:  # dli toggle all
                                                    for t in threading.enumerate():
                                                        if t.name == "pwr_toggle_refresh":
                                                            t.join()  # if refresh thread is running join ~
                                                            # wait for it to complete.
                                                            # TODO Don't think this works or below
                                                            # wouldn't have been necessary.

                                                            # toggle all returns True (ON) or False (OFF) if command
                                                            # successfully sent.  In reality the ports
                                                            # may not be in the  state yet, but dli is working it.
                                                            # Update menu items to reflect end state
                                                            for p in pwr.data["dli_power"][_addr]:
                                                                pwr.data["dli_power"][_addr][p][
                                                                    "state"
                                                                ] = response
                                                            break
                                                self.spin.stop()
                                            # Cycle operation returns False if outlet is off, only valid on powered outlets
                                            elif (
                                                menu_actions[ch]["function"].__name__ == "pwr_cycle"
                                                and not response
                                            ):
                                                log.show(
                                                    f"{host_short} Port {_port} is Off.  Cycle is not valid"
                                                )
                                            elif (
                                                menu_actions[ch]["function"].__name__
                                                == "pwr_rename"
                                            ):
                                                if response:
                                                    _name = pwr._dli[_addr].name(_port)
                                                    if _grp in pwr.data.get("defined", {}):
                                                        pwr.data["defined"][_grp]["is_on"][_port][
                                                            "name"
                                                        ] = _name
                                                    else:
                                                        threading.Thread(
                                                            target=self.outlet_update,
                                                            kwargs={
                                                                "upd_linked": True,
                                                                "refresh": True,
                                                            },
                                                            name="pwr_rename_refresh",
                                                        ).start()
                                                    pwr.data["dli_power"][_addr][_port][
                                                        "name"
                                                    ] = _name
                                        # --// str responses are errors append to error_msgs \\--
                                        # TODO refactor response to use new cpi.response(...)
                                        elif isinstance(response, str) and _port is not None:
                                            log.show(response)
                                        # --// Can Remove After Refactoring all responses to bool or str \\--
                                        elif isinstance(response, int):
                                            if (
                                                menu_actions[ch]["function"].__name__ == "pwr_cycle"
                                                and _port == "all"
                                            ):
                                                if response != 200:
                                                    log.show(
                                                        "Error Response Returned {}".format(
                                                            response
                                                        )
                                                    )
                                            # This is a catch as for the most part I've tried to refactor so the pwr library
                                            # returns port state on success (True/False)
                                            else:
                                                if response in [200, 204]:
                                                    log.show(
                                                        "DEV NOTE: check pwr library ret=200 or 204"
                                                    )
                                                else:
                                                    _action = menu_actions[ch]["function"].__name__
                                                    log.show(
                                                        f"Error returned from dli {host_short} when "
                                                        f"attempting to {_action} port {_port}"
                                                    )
                                    # --// EVAL responses for espHome outlets \\--
                                    elif _type == "esphome":
                                        host_short = utils.get_host_short(_addr)
                                        _port = menu_actions[ch]["kwargs"]["port"]
                                        # --// Operations performed on ALL outlets \\--
                                        if isinstance(response, bool) and _port is not None:
                                            pwr.data["defined"][_grp]["is_on"][_port][
                                                "state"
                                            ] = response
                                            if (
                                                menu_actions[ch]["function"].__name__ == "pwr_cycle"
                                                and not response
                                            ):
                                                _msg = (
                                                    f"{_grp}({host_short})"
                                                    if _grp != host_short
                                                    else f"{_grp}"
                                                )
                                                if _msg != _port:
                                                    _msg = f"{_msg} Port {_port} is Off. Cycle is not valid"
                                                else:
                                                    _msg = f"{_msg} is Off. Cycle is not valid"
                                                log.show(_msg)
                                        elif isinstance(response, str) and _port is not None:
                                            log.show(response)

                                    # --// EVAL responses for GPIO and tasmota outlets \\--
                                    else:
                                        if menu_actions[ch]["function"].__name__ == "pwr_toggle":
                                            if _grp in pwr.data.get("defined", {}):
                                                if isinstance(response, bool):
                                                    pwr.data["defined"][_grp]["is_on"] = response
                                                else:
                                                    pwr.data["defined"][_grp]["errors"] = response
                                        elif (
                                            menu_actions[ch]["function"].__name__ == "pwr_cycle"
                                            and not response
                                        ):
                                            log.show(
                                                "Cycle is not valid for Outlets in the off state"
                                            )
                                        elif menu_actions[ch]["function"].__name__ == "pwr_rename":
                                            log.show(
                                                "rename not yet implemented for {} outlets".format(
                                                    _type
                                                )
                                            )
                            elif calling_menu in ["key_menu", "rename_menu"]:
                                if response:
                                    log.show(response)
                        else:  # not confirmed
                            log.show("Operation Aborted by User")
                elif menu_actions[ch].__name__ in ["power_menu", "dli_menu"]:
                    menu_actions[ch](calling_menu=calling_menu)
                else:  # the selection is a function / coroutine
                    # TODO may need to add inspect.iscoroutinefunction(func) if any of the options are async
                    menu_actions[ch]()
            except KeyError as e:
                if len(choice.orig) <= 2 or not self.exec_shell_cmd(choice.orig):
                    log.show(f"Invalid selection {e}, please try again.")
            except (KeyboardInterrupt, EOFError):
                log.show("Operation Aborted")
                print("")  # prevents header and prompt on same line in debug
        return True

    def confirm_and_spin(self, action_dict: dict, *args, **kwargs):
        """
        called by menu_exec.
        Collects user Confirmation if operation warrants it (Powering off or cycle outlets)
        and Generates appropriate spinner text

        returns tuple
            0: Bool False indicates user abort otherwise True should be returned
            1: str spinner_text used in menu_exec while function runs
            3: str name (for rename operation)
        """
        pwr = self.pwr
        menu = self.menu
        _func = action_dict["function"].__name__
        _off_str = "{{red}}OFF{{norm}}"
        _on_str = "{{green}}ON{{norm}}"
        _cycle_str = "{{red}}C{{green}}Y{{red}}C{{green}}L{{red}}E{{norm}}"
        _type = _addr = None
        to_state = kwargs.get("desired_state")
        if _func in ["pwr_toggle", "pwr_cycle", "pwr_rename"]:
            _type = args[0].lower()
            _addr = args[1]
            _grp = action_dict["key"]
            if _type == "dli":
                port = kwargs["port"]
                if not port == "all":
                    port_name = pwr.data["dli_power"][_addr][port]["name"]
                    to_state = not pwr.data["dli_power"][_addr][port]["state"]
            elif _type == "esphome":
                port = port_name = kwargs["port"]
                if not port == "all":
                    to_state = not pwr.data["defined"][_grp]["is_on"][port]["state"]
            else:
                port = f"{_type}:{_addr}"
                port_name = _grp
                to_state = not pwr.data["defined"][_grp]["is_on"]

        if _type == "dli" or _type == "tasmota" or _type == "esphome":
            host_short = utils.get_host_short(_addr)
        else:
            host_short = None

        prompt = spin_text = name = confirmed = None  # init
        if _func == "pwr_all":
            if kwargs["action"] == "cycle":
                prompt = "{} Power Cycle All Powered {} Outlets".format(
                    "" if _type is None else _type + ":" + host_short, _on_str
                )
                spin_text = "Cycling All{} Ports".format(
                    "" if _type is None else " " + _type + ":" + host_short
                )
            elif kwargs["action"] == "toggle":
                if not kwargs["desired_state"]:
                    prompt = "Power All{} Outlets {}".format(
                        "" if _type is None else " " + _type + ":" + host_short,
                        _off_str,
                    )
                spin_text = "Powering {} ALL{} Outlets".format(
                    menu.format_line(kwargs["desired_state"]).text,
                    "" if _type is None else _type + " :" + host_short,
                )
        elif _func == "pwr_toggle":
            if _type == "dli" and port == "all":
                prompt = "Power {} ALL {} Outlets".format(
                    _off_str if not to_state else _on_str, host_short
                )
            elif not to_state:
                if _type == "dli":
                    prompt = f"Power {_off_str} {host_short} Outlet {port}({port_name})"
                elif _type == "esphome":
                    prompt = f"Power {_off_str} {host_short} Outlet {port}"
                else:  # GPIO or TASMOTA
                    prompt = f"Power {_off_str} Outlet {_grp}({_type}:{_addr})"

            spin_text = "Powering {} {}Outlet{}".format(
                menu.format_line(to_state).text,
                "ALL " if port == "all" else "",
                "s" if port == "all" else "",
            )
        elif _func == "pwr_rename":
            try:
                name = input(
                    "New name for{} Outlet {}: ".format(
                        " " + host_short if host_short else "",
                        port_name if not _type == "dli" else str(port) + "(" + port_name + ")",
                    )
                )
            except (KeyboardInterrupt, EOFError):
                name = None
                confirmed = False
                print("")  # So header doesn't print on same line as aborted prompt when DEBUG is on
            if name:
                old_name = port_name
                _rnm_str = "{red}{old_name}{norm} --> {green}{name}{norm}".format(
                    red="{{red}}",
                    green="{{green}}",
                    norm="{{norm}}",
                    old_name=old_name,
                    name=name,
                )
                if _type == "dli":
                    prompt = "Rename {} Outlet {}: {} ".format(host_short, port, _rnm_str)
                else:
                    old_name = _grp
                    prompt = "Rename Outlet {}:{} {} ".format(_type, host_short, _rnm_str)

                spin_text = "Renaming Port"

        elif _func == "pwr_cycle":
            if _type == "dli" and port == "all":
                prompt = "Power {} ALL {} Outlets".format(_cycle_str, host_short)
            elif _type == "dli":
                prompt = "Cycle Power on {} Outlet {}({})".format(host_short, port, port_name)
            elif _type == "esphome":
                _msg = f"{_grp}({host_short})" if _grp != host_short else f"{_grp}"
                prompt = f"Cycle Power on {_msg} Outlet {port}"
            else:  # GPIO or TASMOTA
                prompt = "Cycle Power on Outlet {}({})".format(port_name, port)
            spin_text = "Cycling {}Outlet{}".format(
                "ALL " if port == "all" else "", "s" if port == "all" else ""
            )

        if prompt:
            prompt = menu.format_line(prompt).text
            confirmed = confirmed if confirmed is not None else utils.user_input_bool(prompt)
        else:
            if _func != "pwr_rename":
                confirmed = True

        return confirmed, spin_text, name
